/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2013 Klaus Spanderen
 
 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.

 compile
export LD_LIBRARY_PATH=/home/klaus/workspace/QuantLib/ql/.libs:/usr/lib/R/lib:/usr/lib/R/library/RInside/lib:/usr/lib/R/library/Rcpp/lib:/usr/lib/R/library/RInside/lib:/usr/lib/R/library/Rcpp/lib

g++ -O3 -I/usr/share/R/include -I/usr/lib/R/include -I/usr/lib/R/library/RInside/include -I/usr/lib/R/library/Rcpp/include -I/home/klaus/workspace/QuantLib -o hestonfpe hestonfpe.cpp -L/home/klaus/workspace/QuantLib/ql/.libs -lQuantLib -L/usr/lib/R/library/RInside/lib -lRInside -L/usr/lib/R/library/Rcpp/lib -lRcpp -L/usr/lib/R/lib -lR

*/


#include <ql/quotes/simplequote.hpp>
#include <ql/processes/hestonprocess.hpp>
#include <ql/time/daycounters/actualactual.hpp>
#include <ql/termstructures/yield/flatforward.hpp>
#include <ql/methods/finitedifferences/schemes/douglasscheme.hpp>
#include <ql/methods/finitedifferences/schemes/hundsdorferscheme.hpp>
#include <ql/methods/finitedifferences/meshers/uniform1dmesher.hpp>
#include <ql/methods/finitedifferences/meshers/fdmmeshercomposite.hpp>
#include <ql/methods/finitedifferences/meshers/fdmblackscholesmesher.hpp>
#include <ql/methods/finitedifferences/operators/fdmlinearoplayout.hpp>
#include <ql/methods/finitedifferences/operators/fdmlinearopcomposite.hpp>
#include <ql/methods/finitedifferences/solvers/fdmbackwardsolver.hpp>
#include <ql/experimental/finitedifferences/fdmhestonfwdop.hpp>

#include <RInside.h>  // for the embedded R via RInside
#include <sstream>

using namespace QuantLib;

std::vector<std::string> rScript(Size i, double t) {
    
    std::stringstream s, d; 
    s << "CairoPNG(\"plot." << i << ".png\") ";
    d << std::fixed << std::setprecision(3) 
      << "main=\"Time: " << t << "Y\")";

    std::string script[] = {
        "if (nchar(installed.packages()[,'Package']['Cairo'])==5) { "
        "  library(\"Cairo\"); ",
            "  if (Cairo.capabilities()['png']) { ",
            s.str(),
        "  } else { "
        "  pdf(\"plot.pdf\") }"
        "} else { pdf(\"plot.pdf\") }",

        "xs <- seq(length(x)*0.3,length(x)/1.6)",
        "ys <- seq(1,length(y)/1.5,10)",
        "x1 <- x[xs]; y1 <- y[ys]; m1 <- m[xs,ys];",
        "persp(x1,y1,m1,theta=140,phi=40,xlab=\"log(S)\","
        "ylab=\"v\",zlab=\"p(S,v,t)\",shade=.25,col=\"lightgray\","
        "zlim=c(0,max(m1)),",
        d.str(),
        "dev.off()"
    };

    return std::vector<std::string>(
        script, script + sizeof(script)/sizeof(std::string));
}

boost::shared_ptr<YieldTermStructure> 
flatRate(const Date& today, Rate forward, const DayCounter& dc) {
    return boost::shared_ptr<YieldTermStructure>(
	new FlatForward(today, Handle<Quote>(
	    boost::shared_ptr<Quote>(new SimpleQuote(forward))), dc));
}

int main(int argc, char *argv[]) {
    const DayCounter dc = ActualActual();
    const Date todaysDate = Date(28, Dec, 2012);
    Settings::instance().evaluationDate() = todaysDate;

    const Date maturityDate = todaysDate + Period(1, Years);
    const Time maturity = dc.yearFraction(todaysDate, maturityDate);

    const Real s0 = 100;
    const Rate r = 0.10;
    const Rate q = 0.05;

    const Real kappa =  2.0;
    const Real theta =  0.4;
    const Real rho   = -0.75;
    const Real sigma =  2;
    const Real v0    =  5*theta;

    const Handle<Quote> spot(boost::shared_ptr<Quote>(new SimpleQuote(s0)));
    const Handle<YieldTermStructure> rTS(flatRate(todaysDate, r, dc));
    const Handle<YieldTermStructure> qTS(flatRate(todaysDate, q, dc));

    boost::shared_ptr<HestonProcess> process(
        new HestonProcess(rTS, qTS, spot, v0, kappa, theta, sigma, rho));

    const Size xGrid = 101;
    const Size vGrid = 1001;
    const Size tGrid = 800;

    const Real vol = sigma*std::sqrt(theta/(2*kappa));
    const Real upperBound = std::max(v0+6*vol, theta+6*vol);
    const Real lowerBound = std::max(0.0025, std::min(v0-6*vol, theta-6*vol));

    const boost::shared_ptr<Fdm1dMesher> varianceMesher(
        new Uniform1dMesher(lowerBound, upperBound, vGrid));
    const boost::shared_ptr<Fdm1dMesher> equityMesher(
        new FdmBlackScholesMesher(
            xGrid,
            FdmBlackScholesMesher::processHelper(
              process->s0(), process->dividendYield(),
              process->riskFreeRate(), std::sqrt(v0)),
              maturity, s0));

    const boost::shared_ptr<FdmMesherComposite>
        mesher(new FdmMesherComposite(equityMesher, varianceMesher));

    Array p(mesher->layout()->size(), 0.0);

    const Size xIdx = xGrid/2;
    const Size vIdx
        = std::distance(varianceMesher->locations().begin(),
                        std::lower_bound(varianceMesher->locations().begin(),
                                         varianceMesher->locations().end(),
                                         v0));
    const Real dx = 0.5*(equityMesher->location(xIdx+1)
                         - equityMesher->location(xIdx-1));
    const Real dy = 0.5*(varianceMesher->location(vIdx+1)
                         - varianceMesher->location(vIdx-1));

    p[xIdx + vIdx*xGrid] = 1.0/(dx*dy);
    Array pd(p.size());

    const boost::shared_ptr<FdmLinearOpComposite> hestonFwdOp(
        new FdmHestonFwdOp(mesher, process));

    HundsdorferScheme evolver(FdmSchemeDesc::Hundsdorfer().theta,
                              FdmSchemeDesc::Hundsdorfer().mu,
                              hestonFwdOp);

    const Time dt = maturity/tGrid;
    evolver.setStep(dt);

    const Size n = equityMesher->size();
    const Size m = varianceMesher->size();

    // embedded R script. Remove the following lines if you don't use RInside
    RInside R(argc, argv);
    Rcpp::NumericVector x(n);
    Rcpp::NumericVector y(m);
    R["x"] = std::vector<Real>(equityMesher->locations().begin(), 
                               equityMesher->locations().end());

    R["y"] = std::vector<Real>(varianceMesher->locations().begin(), 
                               varianceMesher->locations().end());
 
    Rcpp::NumericMatrix M(n, m);
    R["m"] = M;
    
    Size tIdx = 0;
    for (Time t=dt; t <= maturity+20*QL_EPSILON; t+=dt, ++tIdx) {
        evolver.step(p, t);

        for (Size i=0; i < n; ++i) {
            for (Size j=0; j < m; ++j) {
                M(i,j) = p[i+j*n];
            }
        }

        std::vector<std::string> script = rScript(tIdx, t);
        for (Size i=0; i < script.size(); ++i) {
            R.parseEvalQ(script[i]);
        }
    }
}   

